"""Anjani exception"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import traceback
from typing import List, Optional


def format_exception(exp: BaseException, tb: Optional[List[traceback.FrameSummary]] = None) -> str:
    """Formats an exception traceback as a string, similar to the Python interpreter."""

    if tb is None:
        tb = traceback.extract_tb(exp.__traceback__)

    # Replace absolute paths with relative paths
    cwd = os.getcwd()
    for frame in tb:
        if cwd in frame.filename:
            frame.filename = os.path.relpath(frame.filename)

    stack = "".join(traceback.format_list(tb))
    msg = str(exp)
    if msg:
        msg = ": " + msg

    return f"Traceback (most recent call last):\n{stack}{type(exp).__name__}{msg}"



"""Anjani base"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging
from typing import Optional

import aiohttp
import pyrogram

from anjani.util.config import TelegramConfig

from .command_dispatcher import CommandDispatcher
from .database_provider import DatabaseProvider
from .event_dispatcher import EventDispatcher
from .plugin_extenter import PluginExtender
from .telegram_bot import TelegramBot


class Anjani(TelegramBot, DatabaseProvider, PluginExtender, CommandDispatcher, EventDispatcher):
    # Initialized during instantiation
    log: logging.Logger
    http: aiohttp.ClientSession
    client: pyrogram.client.Client
    config: TelegramConfig[str, str]
    loop: asyncio.AbstractEventLoop
    stopping: bool

    def __init__(self, config: TelegramConfig[str, str]):
        self.config = config
        self.log = logging.getLogger("bot")
        self.loop = asyncio.get_event_loop()
        self.stopping = False

        # Initialize mixins
        super().__init__()

        # Initialize aiohttp session last in case another mixin fails
        self.http = aiohttp.ClientSession()

    @classmethod
    async def init_and_run(
        cls, config: TelegramConfig[str, str], *, loop: Optional[asyncio.AbstractEventLoop] = None
    ) -> "Anjani":
        anjani = None

        if loop:
            asyncio.set_event_loop(loop)

        try:
            anjani = cls(config)
            await anjani.run()
            return anjani
        finally:
            asyncio.get_event_loop().stop()

    async def stop(self) -> None:
        self.stopping = True

        self.log.info("Stopping")
        if self.loaded:
            await self.dispatch_event("stop")
            if self.client.is_connected:
                await self.client.stop()

        await self.http.close()
        await self.db.close()

        self.log.info("Running post-stop hooks")
        if self.loaded:
            await self.dispatch_event("stopped")



"""Anjani database collection"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    List,
    Literal,
    Mapping,
    Optional,
    Tuple,
    Union,
)

from bson.codec_options import CodecOptions
from bson.timestamp import Timestamp
from pymongo.collation import Collation
from pymongo.collection import Collection
from pymongo.operations import IndexModel
from pymongo.read_concern import ReadConcern
from pymongo.read_preferences import _ServerMode
from pymongo.results import (
    BulkWriteResult,
    DeleteResult,
    InsertManyResult,
    InsertOneResult,
    UpdateResult,
)
from pymongo.typings import _DocumentType
from pymongo.write_concern import WriteConcern

from anjani import util

from .base import AsyncBaseProperty
from .change_stream import AsyncChangeStream
from .client_session import AsyncClientSession
from .command_cursor import AsyncLatentCommandCursor
from .cursor import AsyncCursor, AsyncRawBatchCursor, Cursor
from .typings import ReadPreferences, Request

if TYPE_CHECKING:
    from .db import AsyncDatabase


class AsyncCollection(AsyncBaseProperty, Generic[_DocumentType]):
    """AsyncIO :obj:`~Collection`

    *DEPRECATED* methods are removed in this class.
    """

    database: "AsyncDatabase"
    dispatch: Collection

    def __init__(
        self,
        database: "AsyncDatabase",
        name: str,
        *,
        create: bool = False,
        codec_options: Optional[CodecOptions] = None,
        read_preference: Optional[_ServerMode] = None,
        write_concern: Optional[WriteConcern] = None,
        read_concern: Optional[ReadConcern] = None,
        collection: Optional[Collection] = None,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> None:
        dispatch = (
            collection
            if collection is not None
            else Collection(
                database.dispatch,
                name,
                create=create,
                codec_options=codec_options,
                read_preference=read_preference,
                write_concern=write_concern,
                read_concern=read_concern,
                session=session.dispatch if session else session,
                **kwargs,
            )
        )
        # Propagate initialization to base
        super().__init__(dispatch)
        self.database = database

    def __bool__(self) -> bool:
        return self.dispatch is not None

    def __getitem__(self, name: str) -> "AsyncCollection":
        return AsyncCollection(
            self.database,
            f"{self.name}.{name}",
            collection=self.dispatch[name],
        )

    def __hash__(self) -> int:
        return hash((self.database, self.name))

    def aggregate(
        self,
        pipeline: List[Mapping[str, Any]],
        *args: Any,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> AsyncLatentCommandCursor:
        return AsyncLatentCommandCursor(
            self,
            self.dispatch.aggregate,
            pipeline,
            session=session.dispatch if session else session,
            *args,
            **kwargs,
        )

    def aggregate_raw_batches(
        self,
        pipeline: List[Mapping[str, Any]],
        *,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> AsyncLatentCommandCursor:
        return AsyncLatentCommandCursor(
            self,
            self.dispatch.aggregate_raw_batches,
            pipeline,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def bulk_write(
        self,
        request: List[Request],
        *,
        ordered: bool = True,
        bypass_document_validation: bool = False,
        session: Optional[AsyncClientSession] = None,
    ) -> BulkWriteResult:
        return await util.run_sync(
            self.dispatch.bulk_write,
            request,
            ordered=ordered,
            bypass_document_validation=bypass_document_validation,
            session=session.dispatch if session else session,
        )

    async def count_documents(
        self,
        query: Mapping[str, Any],
        *,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> int:
        return await util.run_sync(
            self.dispatch.count_documents,
            query,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def create_index(self, keys: Union[str, List[Tuple[str, Any]]], **kwargs: Any) -> str:
        return await util.run_sync(self.dispatch.create_index, keys, **kwargs)

    async def create_indexes(
        self,
        indexes: List[IndexModel],
        *,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> List[str]:
        return await util.run_sync(
            self.dispatch.create_indexes,
            indexes,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def delete_many(
        self,
        query: Mapping[str, Any],
        *,
        collation: Optional[Collation] = None,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
    ) -> DeleteResult:
        return await util.run_sync(
            self.dispatch.delete_many,
            query,
            collation=collation,
            hint=hint,
            session=session.dispatch if session else session,
        )

    async def delete_one(
        self,
        query: Mapping[str, Any],
        *,
        collation: Optional[Collation] = None,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
    ) -> DeleteResult:
        return await util.run_sync(
            self.dispatch.delete_one,
            query,
            collation=collation,
            hint=hint,
            session=session.dispatch if session else session,
        )

    async def distinct(
        self,
        key: str,
        query: Optional[Mapping[str, Any]] = None,
        *,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> List[str]:
        return await util.run_sync(
            self.dispatch.distinct,
            key,
            filter=query,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def drop(self, session: Optional[AsyncClientSession] = None) -> None:
        await util.run_sync(self.dispatch.drop, session=session.dispatch if session else session)

    async def drop_index(
        self,
        index_or_name: Union[str, IndexModel],
        *,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> None:
        await util.run_sync(
            self.dispatch.drop_index,
            index_or_name,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def drop_indexes(self, session: Optional[AsyncClientSession] = None, **kwargs) -> None:
        await util.run_sync(
            self.dispatch.drop_indexes, session=session.dispatch if session else session, **kwargs
        )

    async def estimated_document_count(self, **kwargs: Any) -> int:
        return await util.run_sync(self.dispatch.estimated_document_count, **kwargs)

    def find(self, *args: Any, **kwargs: Any) -> AsyncCursor:
        return AsyncCursor(Cursor(self, *args, **kwargs), self)

    async def find_one(
        self, query: Optional[Mapping[str, Any]], *args: Any, **kwargs: Any
    ) -> Optional[Mapping[str, Any]]:
        return await util.run_sync(self.dispatch.find_one, query, *args, **kwargs)

    async def find_one_and_delete(
        self,
        query: Mapping[str, Any],
        *,
        projection: Optional[Union[List[Any], Mapping[str, Any]]] = None,
        sort: Optional[List[Tuple[str, Any]]] = None,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.find_one_and_delete,
            query,
            projection=projection,
            sort=sort,
            hint=hint,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def find_one_and_replace(
        self,
        query: Mapping[str, Any],
        replacement: Mapping[str, Any],
        *,
        projection: Optional[Union[List[Any], Mapping[str, Any]]] = None,
        sort: Optional[List[Tuple[str, Any]]] = None,
        upsert: bool = False,
        return_document: bool = False,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.find_one_and_replace,
            query,
            replacement,
            projection=projection,
            sort=sort,
            upsert=upsert,
            return_document=return_document,
            hint=hint,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def find_one_and_update(
        self,
        query: Mapping[str, Any],
        update: Mapping[str, Any],
        *,
        projection: Optional[Union[List[Any], Mapping[str, Any]]] = None,
        sort: Optional[List[Tuple[str, Any]]] = None,
        upsert: bool = False,
        return_document: bool = False,
        array_filters: Optional[List[Mapping[str, Any]]] = None,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
        **kwargs: Any,
    ) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.find_one_and_update,
            query,
            update,
            projection=projection,
            sort=sort,
            upsert=upsert,
            return_document=return_document,
            array_filters=array_filters,
            hint=hint,
            session=session.dispatch if session else session,
            **kwargs,
        )

    def find_raw_batches(self, *args: Any, **kwargs: Any) -> AsyncRawBatchCursor:
        if "session" in kwargs:
            session = kwargs["session"]
            kwargs["session"] = session.dispatch if session else session

        cursor = self.dispatch.find_raw_batches(*args, **kwargs)

        return AsyncRawBatchCursor(cursor, self)

    async def index_information(
        self, session: Optional[AsyncClientSession] = None
    ) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.index_information, session=session.dispatch if session else session
        )

    async def insert_many(
        self,
        documents: List[Mapping[str, Any]],
        *,
        ordered: bool = True,
        bypass_document_validation: bool = False,
        session: Optional[AsyncClientSession] = None,
    ) -> InsertManyResult:
        return await util.run_sync(
            self.dispatch.insert_many,
            documents,
            ordered=ordered,
            bypass_document_validation=bypass_document_validation,
            session=session.dispatch if session else session,
        )

    async def insert_one(
        self,
        document: Mapping[str, Any],
        *,
        bypass_document_validation: bool = False,
        session: Optional[AsyncClientSession] = None,
    ) -> InsertOneResult:
        return await util.run_sync(
            self.dispatch.insert_one,
            document,
            bypass_document_validation=bypass_document_validation,
            session=session.dispatch if session else session,
        )

    def list_indexes(
        self, *args: Any, session: Optional[AsyncClientSession] = None, **kwargs: Any
    ) -> AsyncLatentCommandCursor:
        return AsyncLatentCommandCursor(
            self,
            self.dispatch.list_indexes,
            session=session.dispatch if session else session,
            *args,
            **kwargs,
        )

    async def options(self, session: Optional[AsyncClientSession] = None) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.options, session=session.dispatch if session else session
        )

    async def rename(
        self, new_name: str, *, session: Optional[AsyncClientSession] = None, **kwargs: Any
    ) -> Mapping[str, Any]:
        return await util.run_sync(
            self.dispatch.rename,
            new_name,
            session=session.dispatch if session else session,
            **kwargs,
        )

    async def replace_one(
        self,
        query: Mapping[str, Any],
        replacement: Mapping[str, Any],
        *,
        upsert: bool = False,
        bypass_document_validation: bool = False,
        collation: Optional[Collation] = None,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
    ) -> UpdateResult:
        return await util.run_sync(
            self.dispatch.replace_one,
            query,
            replacement,
            upsert=upsert,
            bypass_document_validation=bypass_document_validation,
            collation=collation,
            hint=hint,
            session=session.dispatch if session else session,
        )

    async def update_many(
        self,
        query: Mapping[str, Any],
        update: Mapping[str, Any],
        *,
        upsert: bool = False,
        array_filters: Optional[List[Mapping[str, Any]]] = None,
        bypass_document_validation: bool = False,
        collation: Optional[Collation] = None,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
    ) -> UpdateResult:
        return await util.run_sync(
            self.dispatch.update_many,
            query,
            update,
            upsert=upsert,
            array_filters=array_filters,
            bypass_document_validation=bypass_document_validation,
            collation=collation,
            hint=hint,
            session=session.dispatch if session else session,
        )

    async def update_one(
        self,
        query: Mapping[str, Any],
        update: Mapping[str, Any],
        *,
        upsert: bool = False,
        array_filters: Optional[List[Mapping[str, Any]]] = None,
        bypass_document_validation: bool = False,
        collation: Optional[Collation] = None,
        hint: Optional[Union[IndexModel, List[Tuple[str, Any]]]] = None,
        session: Optional[AsyncClientSession] = None,
    ) -> UpdateResult:
        return await util.run_sync(
            self.dispatch.update_one,
            query,
            update,
            upsert=upsert,
            array_filters=array_filters,
            bypass_document_validation=bypass_document_validation,
            collation=collation,
            hint=hint,
            session=session.dispatch if session else session,
        )

    def watch(
        self,
        pipeline: Optional[List[Mapping[str, Any]]] = None,
        *,
        full_document: Optional[Literal["updateLookup"]] = None,
        resume_after: Optional[Mapping[str, str]] = None,
        max_await_time_ms: Optional[int] = None,
        batch_size: Optional[int] = None,
        collation: Optional[Collation] = None,
        start_at_operation_time: Optional[Timestamp] = None,
        session: Optional[AsyncClientSession] = None,
        start_after: Optional[Mapping[str, str]] = None,
        comment: Optional[str] = None,
        full_document_before_change: Optional[Literal["required", "whenAvailable"]] = None,
    ) -> AsyncChangeStream:
        return AsyncChangeStream(
            self,
            pipeline,
            full_document,
            resume_after,
            max_await_time_ms,
            batch_size,
            collation,
            start_at_operation_time,
            session,
            start_after,
            comment,
            full_document_before_change,
        )

    def with_options(
        self,
        *,
        codec_options: Optional[CodecOptions] = None,
        read_preference: Optional[ReadPreferences] = None,
        write_concern: Optional[WriteConcern] = None,
        read_concern: Optional[ReadConcern] = None,
    ) -> "AsyncCollection":
        self.dispatch = self.dispatch.with_options(
            codec_options=codec_options,
            read_preference=read_preference,
            write_concern=write_concern,
            read_concern=read_concern,
        )

        return self

    @property
    def full_name(self) -> str:
        return self.dispatch.full_name

    @property
    def name(self) -> str:
        return self.dispatch.name



"""Anjani command dispatcher"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
from typing import TYPE_CHECKING, Any, Iterable, MutableMapping, Optional, Union

from pyrogram import ContinuePropagation, errors
from pyrogram.client import Client
from pyrogram.enums.chat_action import ChatAction
from pyrogram.enums.chat_type import ChatType
from pyrogram.filters import Filter, create
from pyrogram.types import Message

from anjani import command, plugin, util
from anjani.error import CommandHandlerError, CommandInvokeError, ExistingCommandError
from anjani.util.cache_limiter import CacheLimiter

from .anjani_mixin_base import MixinBase

if TYPE_CHECKING:
    from .anjani_bot import Anjani


class CommandDispatcher(MixinBase):
    # Initialized during instantiation
    commands: MutableMapping[str, command.Command]

    # Private
    __limiter: CacheLimiter

    def __init__(self: "Anjani", **kwargs: Any) -> None:
        # Initialize command map
        self.commands = {}

        self.__limiter = CacheLimiter(ttl=10, max_value=3)

        # Propagate initialization to other mixins
        super().__init__(**kwargs)

    def register_command(
        self: "Anjani",
        plug: plugin.Plugin,
        name: str,
        func: command.CommandFunc,
        *,
        filters: Optional[Union[util.types.CustomFilter, Filter]] = None,
        aliases: Iterable[str] = [],
    ) -> None:
        if getattr(func, "_listener_filters", None):
            self.log.warning(
                "@listener.filters decorator only for ListenerFunc. Filters will be ignored..."
            )

        if filters:
            self.log.debug("Registering filter '%s' into '%s'", type(filters).__name__, name)
            util.misc.check_filters(filters, self)

        cmd = command.Command(name, plug, func, filters, aliases)

        if name in self.commands:
            orig = self.commands[name]
            raise ExistingCommandError(orig, cmd)

        self.commands[name] = cmd

        for alias in cmd.aliases:
            if alias in self.commands:
                orig = self.commands[alias]
                raise ExistingCommandError(orig, cmd, alias=True)

            self.commands[alias] = cmd

    def unregister_command(self: "Anjani", cmd: command.Command) -> None:
        del self.commands[cmd.name]

        for alias in cmd.aliases:
            try:
                del self.commands[alias]
            except KeyError:
                continue

    def register_commands(self: "Anjani", plug: plugin.Plugin) -> None:
        for name, func in util.misc.find_prefixed_funcs(plug, "cmd_"):
            done = False

            try:
                self.register_command(
                    plug,
                    name,
                    func,
                    filters=getattr(func, "_cmd_filters", None),
                    aliases=getattr(func, "_cmd_aliases", []),
                )
                done = True
            finally:
                if not done:
                    self.unregister_commands(plug)

    def unregister_commands(self: "Anjani", plug: plugin.Plugin) -> None:
        # Can't unregister while iterating, so collect commands to unregister afterwards
        to_unreg = []

        for name, cmd in self.commands.items():
            # Let unregister_command deal with aliases
            if name != cmd.name:
                continue

            if cmd.plugin == plug:
                to_unreg.append(cmd)

        # Actually unregister the commands
        for cmd in to_unreg:
            self.unregister_command(cmd)

    def command_predicate(self: "Anjani") -> Filter:
        async def func(_: Filter, client: Client, message: Message) -> bool:
            if message.via_bot:
                return False

            if (message.chat and message.chat.type == ChatType.CHANNEL) or (
                message.sender_chat
                and message.forward_from_chat
                and message.forward_from_chat.id == message.sender_chat.id
            ):
                return False  # ignore channel broadcasts

            if message.text is not None and message.text.startswith("/"):
                user = message.from_user or message.sender_chat
                # check user limiter
                if await self.__limiter.exceeded(user.id):
                    return False

                try:
                    parts = message.text.split()
                    parts[0] = parts[0][1:]

                    # Check if bot command contains a valid username
                    # eg: /ping@dAnjani_bot will return True
                    # If current bot instance is dAnjani_bot else False
                    if self.user.username and self.user.username in parts[0]:
                        # Remove username from command
                        parts[0] = parts[0].replace(f"@{self.user.username}", "")

                    # Filter if command is not in commands
                    try:
                        cmd = self.commands[parts[0]]
                    except KeyError:
                        return False

                    # Check additional built-in filters
                    if cmd.filters:
                        if inspect.iscoroutinefunction(cmd.filters.__call__):
                            if not await cmd.filters(client, message):
                                return False
                        else:
                            if not await util.run_sync(cmd.filters, client, message):
                                return False

                    message.command = parts
                    return True
                finally:
                    # Increment user limit
                    await self.__limiter.increment(user.id)

            return False

        return create(func, "CustomCommandFilter")

    async def on_command(
        self: "Anjani", client: Client, message: Message  # skipcq: PYL-W0613
    ) -> None:
        # cmd never raises KeyError because we checked on command_predicate
        cmd = self.commands[message.command[0]]
        try:
            # Construct invocation context
            ctx = command.Context(
                self,
                message,
                1 + len(message.command[0]) + 1,
            )

            # Parse and convert handler required parameters
            signature = inspect.signature(cmd.func)
            args = []  # type: list[Any]
            kwargs = {}  # type: MutableMapping[str, Any]
            if len(signature.parameters) > 1:
                args, kwargs = await util.converter.parse_arguments(signature, ctx, cmd.func)

            # Invoke command function
            try:
                ret = await cmd.func(ctx, *args, **kwargs)

                # Response shortcut
                if ret is not None:
                    async with ctx.action(ChatAction.TYPING):
                        await ctx.respond(
                            ret,
                            disable_web_page_preview=True,
                        )
            except errors.MessageNotModified:
                cmd.plugin.log.warning(
                    "Command '%s' triggered a message edit with no changes; make sure there is only a single bot instance running",
                    cmd.name,
                )
            except Exception as e:  # skipcq: PYL-W0703
                constructor_invoke = CommandInvokeError(
                    f"raised from {type(e).__name__}: {str(e)}"
                ).with_traceback(e.__traceback__)
                chat = ctx.chat
                user = ctx.msg.from_user
                cmd.plugin.log.error(
                    "Error in command '%s'\n"
                    "  Data:\n"
                    "    • Chat    -> %s (%d)\n"
                    "    • Invoker -> %s (%d)\n"
                    "    • Input   -> %s\n",
                    cmd.name,
                    chat.title if chat else None,
                    chat.id if chat else None,
                    user.first_name if user else None,
                    user.id if user else None,
                    ctx.input,
                    exc_info=constructor_invoke,
                )

            await self.dispatch_event("command", ctx, cmd)
        except Exception as e:  # skipcq: PYL-W0703
            constructor_handler = CommandHandlerError(
                f"raised from {type(e).__name__}: {str(e)}"
            ).with_traceback(e.__traceback__)
            chat = message.chat
            user = message.from_user
            cmd.plugin.log.error(
                "Error in command handler\n"
                "  Data:\n"
                "    • Chat    -> %s (%d)\n"
                "    • Invoker -> %s (%d)\n"
                "    • Input   -> %s\n",
                cmd.name,
                chat.title if chat else None,
                chat.id if chat else None,
                user.first_name if user else None,
                user.id if user else None,
                message.command,
                exc_info=constructor_handler,
            )
        finally:
            # Continue processing handler of on_message
            raise ContinuePropagation



"""Anjani config"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Iterator, MutableMapping, Optional, TypeVar

_KT = TypeVar("_KT", bound=str, contravariant=True)
_VT = TypeVar("_VT", covariant=True)


class TelegramConfig(MutableMapping[_KT, _VT]):
    def __init__(self, config: MutableMapping[str, Any]) -> None:
        for key, value in config.items():
            if not value:
                continue

            super().__setattr__(key, value)

    def is_plugin_disabled(self, name: str) -> bool:
        return name in self.__getattribute__("plugin_flag")

    def __contains__(self, k: _KT) -> bool:
        return k in self.__dict__

    def __delattr__(self, obj: object) -> None:  # skipcq: PYL-W0613
        raise RuntimeError("Can't delete configuration while running the bot.")

    def __delitem__(self, k: _KT) -> None:  # skipcq: PYL-W0613
        raise RuntimeError("Can't delete configuration while running the bot.")

    def __getattr__(self, name: str) -> _VT:
        return self.__getattribute__(name)

    def __getitem__(self, k: _KT) -> _VT:
        return self.__dict__[k]

    def __iter__(self) -> Iterator[str]:
        return self.__dict__.__iter__()

    def __len__(self) -> int:
        return len(self.__dict__)

    def __setattr__(self, name: str, value: Any) -> None:  # skipcq: PYL-W0613
        raise RuntimeError("Configuration must be done before running the bot.")

    def __setitem__(self, k: str, v: Any) -> None:  # skipcq: PYL-W0613
        raise RuntimeError("Configuration must be done before running the bot.")



""" Restriction Plugin. """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import Any, ClassVar, MutableMapping, Optional, Union

from bson.objectid import ObjectId
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.errors import BadRequest, PeerIdInvalid, UserNotParticipant
from pyrogram.types import (
    CallbackQuery,
    Chat,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    Message,
    User,
)

from anjani import command, filters, listener, plugin, util


class Restrictions(plugin.Plugin):
    name: ClassVar[str] = "Restriction"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("CHATS")

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        data = await self.db.find_one(
            {"chat_id": chat_id, "warn_list": {"$exists": True}}, {"_id": 0, "warn_list": 1}
        )
        try:
            return {self.name: data["warn_list"]} if data else {}
        except KeyError:
            return {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one(
            {"chat_id": chat_id}, {"$set": {"warn_list": data[self.name]}}, upsert=True
        )

    @listener.filters(filters.regex(r"^rm_warn_(\d+)_(.*)"))
    async def on_callback_query(self, query: CallbackQuery) -> None:
        """Remove warn callback data"""
        chat = query.message.chat
        user = query.matches[0].group(1)
        uid = query.matches[0].group(2)

        try:
            invoker = await chat.get_member(query.from_user.id)
        except UserNotParticipant:
            return await query.answer(
                await self.get_text(chat.id, "error-no-rights"), show_alert=True
            )

        if not invoker.privileges or not invoker.privileges.can_restrict_members:
            return await query.answer(await self.get_text(chat.id, "warn-keyboard-not-admins"))

        chat_data = await self.db.find_one(
            {"chat_id": chat.id, "warn_list": {"$exists": True}}, {"warn_list": 1}
        )
        if not chat_data:
            return

        warn_list = chat_data.get("warn_list", {})
        try:
            warns = len(warn_list[user])
        except KeyError:
            warns = 0

        if user not in warn_list or warns == 0:
            await query.message.edit(await self.get_text(chat.id, "warn-keyboard-removed"))
            return

        target = await self.bot.client.get_users(int(user))
        if isinstance(target, list):
            target = target[0]

        await asyncio.gather(
            self.db.update_one(
                {"chat_id": chat.id},
                {"$unset": {f"warn_list.{user}.{uid}": ""}},
            ),
            query.message.edit(
                await self.get_text(
                    chat.id,
                    "warn-removed",
                    query.from_user.mention,
                    target.mention,
                )
            ),
        )

    @command.filters(filters.can_restrict)
    async def cmd_kick(
        self, ctx: command.Context, target: Union[User, Chat, None] = None, *, reason: str = ""
    ) -> str:
        """Kick chat member"""
        chat = ctx.chat
        reply_msg = ctx.msg.reply_to_message

        if not target:
            if ctx.args and not reply_msg:
                return await self.text(chat.id, "err-peer-invalid")

            if not reply_msg or not (reply_msg.from_user or reply_msg.sender_chat):
                return await self.text(chat.id, "no-kick-user")

            target = reply_msg.from_user or reply_msg.sender_chat
            reason = ctx.input

        try:
            if isinstance(target, User) and util.tg.is_staff_or_admin(
                await chat.get_member(target.id)
            ):
                return await self.text(chat.id, "admin-kick")
        except UserNotParticipant:
            if util.tg.is_staff(target.id):
                return await self.text(chat.id, "admin-kick")

        await self.bot.client.ban_chat_member(chat.id, target.id)

        async def unban():
            await asyncio.sleep(5)
            await self.bot.client.unban_chat_member(chat.id, target.id)

        asyncio.create_task(unban())

        ret = await self.text(
            chat.id, "kick-done", target.first_name if isinstance(target, User) else target.title
        )
        if reason:
            ret += await self.text(chat.id, "kick-reason", reason)

        return ret

    async def _ban(
        self,
        ctx: command.Context,
        target: Union[User, Chat, None] = None,
        reason: str = "",
        silent: bool = False,
    ) -> Optional[str]:
        chat = ctx.chat
        reply_msg = ctx.msg.reply_to_message

        if not target:
            if ctx.args and not reply_msg:
                return await self.text(chat.id, "err-peer-invalid")

            if not reply_msg or not (reply_msg.from_user or reply_msg.sender_chat):
                return await self.text(chat.id, "no-ban-user")

            target = reply_msg.from_user or reply_msg.sender_chat
            reason = ctx.input

        try:
            if isinstance(target, User) and util.tg.is_staff_or_admin(
                await chat.get_member(target.id)
            ):
                return await self.text(chat.id, "admin-ban")
        except UserNotParticipant:
            # Not a participant in the chat (replying from channel discussion)
            if util.tg.is_staff(target.id):
                return await self.text(chat.id, "admin-ban")

        ret, _ = await asyncio.gather(
            self.text(
                chat.id, "ban-done", target.first_name if isinstance(target, User) else target.title
            ),
            chat.ban_member(target.id),
        )
        if reason:
            ret += await self.text(chat.id, "ban-reason", reason)

        return ret if not silent else None

    @command.filters(filters.can_restrict)
    async def cmd_ban(
        self, ctx: command.Context, target: Union[User, Chat, None] = None, *, reason: str = ""
    ) -> str:
        """Ban chat member"""
        ret = await self._ban(ctx, target, reason)
        if not ret:
            raise ValueError

        return ret

    @command.filters(filters.can_restrict)
    async def cmd_sban(self, ctx: command.Context, target: Union[User, Chat, None] = None) -> None:
        """Silently ban chat member"""
        ret = await self._ban(ctx, target, silent=True)
        if ret:
            await ctx.respond(ret, delete_after=1)

        await ctx.msg.delete()

    @command.filters(filters.can_restrict)
    async def cmd_unban(self, ctx: command.Context, user: Union[User, Chat, None] = None) -> str:
        """Unban chat member"""
        chat = ctx.chat

        if not user:
            if ctx.input:
                return await self.text(chat.id, "err-peer-invalid")

            if not ctx.msg.reply_to_message:
                return await self.text(chat.id, "unban-no-user")

            user = ctx.msg.reply_to_message.from_user or ctx.msg.reply_to_message.sender_chat

        try:
            await chat.unban_member(user.id)
        except (BadRequest, PeerIdInvalid) as e:
            if isinstance(e.value, str) and "PARTICIPANT_ID_INVALID" in e.value:
                return await self.text(chat.id, "err-invalid-pid")

            if isinstance(e, PeerIdInvalid):
                return await self.text(chat.id, "err-peer-invalid")

            raise e from BadRequest

        return await self.text(
            chat.id, "unban-done", user.first_name if isinstance(user, User) else user.title
        )

    @command.filters(filters.can_restrict)
    async def cmd_warn(
        self, ctx: command.Context, user: Optional[User] = None, *, reason: str = ""
    ) -> Optional[str]:
        """Warn command chat member"""
        chat = ctx.chat
        reply_msg = ctx.msg.reply_to_message
        if not user:
            if ctx.input and not reply_msg:
                return await ctx.get_text("err-peer-invalid")

            if not reply_msg:
                return await ctx.get_text("warn-no-user")

            reason = ctx.input
            user = reply_msg.from_user

        if user.id == self.bot.uid:
            return await ctx.get_text("error-its-myself")

        target = await chat.get_member(user.id)
        if target.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}:
            return await ctx.get_text("rmwarn-admin")

        threshold = 3
        warns: Optional[int] = None
        chat_data = await self.db.find_one(
            {"chat_id": chat.id, "warn_list": {"$exists": True}},
            {"warn_list": 1, "warn_threshold": 1},
        )
        if chat_data:  # Get threshold and existing warns from chat data
            threshold = chat_data.get("warn_threshold", 3)
            try:
                warns = len(chat_data["warn_list"][str(user.id)])
            except KeyError:
                pass

        uid = str(ObjectId())
        reason = reason or await ctx.get_text("warn-default-reason")
        keyboard = [
            [
                InlineKeyboardButton(
                    await ctx.get_text("warn-keyboard-text"),
                    callback_data=f"rm_warn_{user.id}_{uid}",
                )
            ]
        ]
        await asyncio.gather(
            self.db.update_one(
                {"chat_id": chat.id}, {"$set": {f"warn_list.{user.id}.{uid}": reason}}, upsert=True
            ),
            ctx.respond(
                await self.get_text(
                    chat.id,
                    "warn-message",
                    user.mention,
                    1 if warns is None else warns + 1,
                    threshold,
                    reason,
                ),
                reply_markup=InlineKeyboardMarkup(keyboard),
            ),
        )

        if warns is not None and warns + 1 >= threshold:
            ret, _ = await asyncio.gather(
                ctx.get_text("warn-user-max", user.mention), chat.ban_member(user.id)
            )
            return ret

        return None

    @command.filters(filters.group)
    async def cmd_warns(self, ctx: command.Context) -> str:
        """Get warns of a user"""
        reply_msg = ctx.msg.reply_to_message
        chat, user = ctx.chat, reply_msg.from_user if reply_msg else ctx.author
        threshold = 3

        if user.id == self.bot.uid:
            return await ctx.get_text("error-its-myself")

        target = await chat.get_member(user.id)
        if target.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}:
            return await ctx.get_text("rmwarn-admin")

        chat_data = await self.db.find_one(
            {"chat_id": chat.id, "warn_list": {"$exists": True}},
            {"warn_list": 1, "warn_threshold": 1},
        )
        if not chat_data:
            return await ctx.get_text("warn-no-data", user.mention)
        threshold = chat_data.get("warn_threshold", 3)

        try:
            warns_list = chat_data["warn_list"][str(user.id)]
        except KeyError:
            return await ctx.get_text("warn-no-data", user.mention)
        else:
            warns = len(warns_list)
            if warns == 0:
                return await ctx.get_text("warn-no-data", user.mention)

        return await self.get_text(
            chat.id,
            "warn-message-list",
            user.mention,
            warns,
            threshold,
            "\n".join(f"  • __{reason}__" for reason in warns_list.values()),
        )

    @command.filters(filters.admin_only, aliases={"warnlim"})
    async def cmd_warnlimit(self, ctx: command.Context, limit: int) -> Optional[str]:
        """Set warn limit"""
        chat = ctx.chat
        if limit is None:
            return await ctx.get_text("warn-limit-no-input")

        if not isinstance(limit, int):
            return await ctx.get_text("warn-limit-invalid-input")

        if limit <= 0:
            return await ctx.get_text("warn-limit-invalid-input")

        ret, _ = await asyncio.gather(
            ctx.get_text("warn-limit-done", limit),
            self.db.update_one({"chat_id": chat.id}, {"$set": {"warn_threshold": limit}}),
        )
        return ret

    @command.filters(filters.can_restrict, aliases={"removewarn"})
    async def cmd_rmwarn(self, ctx: command.Context, user: Optional[User] = None) -> Optional[str]:
        """Remove a warn"""
        chat = ctx.chat
        reply_msg = ctx.msg.reply_to_message

        if not user:
            if ctx.input and not reply_msg:
                return await ctx.get_text("err-peer-invalid")

            if not reply_msg:
                return await ctx.get_text("rmwarn-no-user")

            user = reply_msg.from_user

        if user.id == self.bot.uid:
            return await ctx.get_text("error-its-myself")

        target = await chat.get_member(user.id)
        if target.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}:
            return await ctx.get_text("rmwarn-admin")

        chat_data = await self.db.find_one(
            {"chat_id": chat.id, "warn_list": {"$exists": True}}, {"warn_list": 1}
        )
        if not chat_data:
            return await ctx.get_text("warn-no-data", user.mention)

        try:
            warns_list = chat_data["warn_list"][str(user.id)]
        except KeyError:
            return await ctx.get_text("warn-no-data", user.mention)

        try:
            ret, _ = await asyncio.gather(
                self.get_text(chat.id, "rmwarn-done", user.mention),
                self.db.update_one(
                    {"chat_id": chat.id},
                    {"$unset": {f"warn_list.{user.id}.{list(warns_list.keys())[-1]}": ""}},
                ),
            )
        except IndexError:
            return await ctx.get_text("warn-no-data", user.mention)

        return ret



"""Anjani database core"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from typing import TYPE_CHECKING, Any

from anjani import util

from .anjani_mixin_base import MixinBase

if TYPE_CHECKING:
    from .anjani_bot import Anjani


class DatabaseProvider(MixinBase):
    db: util.db.AsyncDatabase

    def __init__(self: "Anjani", **kwargs: Any) -> None:
        if sys.platform == "win32":
            import certifi

            client = util.db.AsyncClient(
                self.config["db_uri"], connect=False, tlsCAFile=certifi.where()
            )
        else:
            client = util.db.AsyncClient(self.config["db_uri"], connect=False)

        self.db = client.get_database("AnjaniBot")

        # Propagate initialization to other mixins
        super().__init__(**kwargs)



"""Chat SpamShield"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from datetime import datetime
from json import JSONDecodeError
from typing import Any, ClassVar, List, MutableMapping, Optional

from aiohttp import (
    ClientConnectorError,
    ClientOSError,
    ClientResponseError,
    ContentTypeError,
)
from pyrogram.errors import (
    BadRequest,
    ChannelPrivate,
    ChatAdminRequired,
    PeerIdInvalid,
    UserNotParticipant,
)
from pyrogram.types import Chat, Message, User

try:
    from userbotindo import get_trust
except ImportError:
    from anjani.util.misc import do_nothing as get_trust

from anjani import command, filters, listener, plugin, util
from anjani.util.misc import StopPropagation


class SpamShield(plugin.Plugin):
    name: ClassVar[str] = "SpamShield"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection
    federation_db: util.db.AsyncCollection
    token: Optional[str]
    spam_protection: bool

    async def on_load(self) -> None:
        self.token = self.bot.config.get("sw_api")
        if not self.token:
            self.bot.log.warning("SpamWatch API token not exist")

        self.db = self.bot.db.get_collection("GBAN_SETTINGS")  # spamshield autoban
        self.federation_db = self.bot.db.get_collection("FEDERATIONS")
        self.user_db = self.bot.db.get_collection("USERS")
        self.spam_protection = "SpamPredict" in self.bot.plugins

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        setting = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: setting} if setting else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    @listener.priority(90)
    async def on_chat_action(self, message: Message) -> None:
        """Checker service for new member"""
        chat = message.chat
        if message.left_chat_member or not await self.is_active(chat.id):
            return

        try:
            me = await chat.get_member("me")
            if not me.privileges or not me.privileges.can_restrict_members:
                return

            tasks = set()
            for member in message.new_chat_members:
                tasks.add(self.bot.loop.create_task(self.check(member, chat)))

            res: List[bool] = await asyncio.gather(*tasks)
            # Assume only one member is added at a time, so raise StopPropagation
            if all(res):
                raise StopPropagation
        except ChannelPrivate:
            return

    @listener.priority(65)
    @listener.filters(filters.group & ~filters.outgoing)
    async def on_message(self, message: Message) -> None:
        """Checker service for message"""
        chat = message.chat
        user = message.from_user
        text = (
            message.text.strip()
            if message.text
            else (message.caption.strip() if message.media and message.caption else None)
        )
        if not chat or not user or not text or not await self.is_active(chat.id):
            return

        if self.spam_protection:
            sample = await self.user_db.find_one({"_id": user.id}, {"pred_sample": 1, "spam": 1})
            if sample and not sample.get("spam", False):
                trust = get_trust(sample.get("pred_sample", []))
                if trust and trust < 5.0:
                    self.log.debug(f"{user.id} has low trust score, flaging as spam")
                    await self.user_db.update_one({"_id": user.id}, {"$set": {"spam": True}})

        try:
            me, target = await util.tg.fetch_permissions(self.bot.client, chat.id, user.id)
            if (
                not (me and target)
                or not me.privileges
                or not me.privileges.can_restrict_members
                or util.tg.is_staff_or_admin(target)
            ):
                return

            await self.check(target.user, chat)
        except (ChannelPrivate, ChatAdminRequired, PeerIdInvalid, UserNotParticipant):
            return

    async def get_ban(self, user_id: int) -> MutableMapping[str, Any]:
        if not self.token:
            return {}

        path = f"https://api.spamwat.ch/banlist/{user_id}"
        headers = {"Authorization": f"Bearer {self.token}"}
        try:
            async with self.bot.http.get(path, headers=headers) as resp:
                if resp.status in {200, 201}:
                    return await resp.json()

                if resp.status == 404:
                    return {}

                if resp.status == 401:
                    self.log.error(
                        "Spamwatch API error",
                        exc_info=ClientResponseError(
                            resp.request_info,
                            resp.history,
                            message="Make sure your Spamwatch API token is corret",
                        ),
                    )
                    return {}

                if resp.status == 403:
                    self.log.error(
                        "Spamwatch API error",
                        exc_info=ClientResponseError(
                            resp.request_info,
                            resp.history,
                            message="Forbidden, your token permissions is not valid",
                        ),
                    )
                    return {}

                if resp.status == 429:
                    self.log.warning(
                        "Spamwatch API error",
                        exc_info=ClientResponseError(
                            resp.request_info,
                            resp.history,
                            message="There were problems with request... Too many.",
                        ),
                    )
                    return {}

                self.log.error(
                    f"Unknown Spamwatch API error: Received {resp.status}",
                    exc_info=ClientResponseError(resp.request_info, resp.history),
                )
                return {}
        except ClientConnectorError:
            return {}

    async def cas_check(self, user: User) -> Optional[str]:
        """Check on CAS"""
        retry = 0
        while True:
            try:
                async with self.bot.http.get(
                    f"https://api.cas.chat/check?user_id={user.id}"
                ) as res:
                    data = await res.json()
                    if data["ok"]:
                        reason = f"https://cas.chat/query?u={user.id}"
                        return reason

                    return None
            except (ContentTypeError, JSONDecodeError):
                if retry == 5:
                    self.log.debug("Error parsing CAS response")
                    return None

                retry += 1
                await asyncio.sleep(1)
                self.log.debug("Invalid data received from CAS server, retrying...")
            except ClientOSError:
                if retry == 10:
                    self.log.debug("Error connecting to CAS API")
                    return None

                retry += 1
                await asyncio.sleep(0.5)
                self.log.debug(f"Retrying CAS check for {user.id}")

    async def check_spam(self, uid: int) -> bool:
        if not self.spam_protection:
            return False
        res = await self.user_db.find_one({"_id": uid}, {"spam": 1})
        return res.get("spam", False) if res else False

    async def is_active(self, chat_id: int) -> bool:
        """Return SpamShield setting"""
        data = await self.db.find_one({"chat_id": chat_id})
        return data["setting"] if data else True

    async def ban(self, chat: Chat, user: User, reason: str) -> None:
        fullname = user.first_name + user.last_name if user.last_name else user.first_name
        await asyncio.gather(
            chat.ban_member(user.id),
            self.federation_db.update_one(
                {"_id": "AnjaniSpamShield"},
                {
                    "$set": {
                        f"banned.{user.id}": {
                            "name": fullname,
                            "reason": "Automated fban " + reason,
                            "time": datetime.now(),
                        }
                    }
                },
            ),
        )

    async def setting(self, chat_id: int, setting: bool) -> None:
        """Turn on/off SpamShield in chats"""
        if setting:
            await self.db.update_one({"chat_id": chat_id}, {"$set": {"setting": True}}, upsert=True)
        else:
            await self.db.delete_one({"chat_id": chat_id})

    async def check(self, user: User, chat: Chat) -> bool:
        """Shield checker action."""
        cas, sw, spam = await asyncio.gather(
            self.cas_check(user), self.get_ban(user.id), self.check_spam(user.id)
        )
        if not (cas or sw or spam):
            return False

        userlink = user.mention
        chat_link = f"[{chat.id}](https://t.me/{chat.username})" if chat.username else str(chat.id)
        reason = ""
        banner = ""
        if cas:
            banner = "[Combot Anti Spam](t.me/combot)"
            reason = f"[Link]({cas})"
        if sw:
            if not banner:
                banner = "[Spam Watch](t.me/SpamWatch)"
                reason = sw["reason"]
            else:
                banner += " & [Spam Watch](t.me/SpamWatch)"
                reason += " & " + sw["reason"]
        if user.is_scam:  # overwrite banner and reason if user is flagged by telegram
            banner = "Telegram Server"
            reason = "Flagged as a scammer."
        if spam:
            banner = "Anjani Spam Protection"
            reason = "Flagged as a spammer."

        try:
            await asyncio.gather(
                self.bot.log_stat("banned"),
                self.ban(chat, user, reason),
                self.bot.client.send_message(
                    chat.id,
                    text=await self.text(chat.id, "banned-text", userlink, user.id, reason, banner),
                    disable_web_page_preview=True,
                ),
                self.bot.client.send_message(
                    int(self.bot.config.log_channel),
                    text=(
                        "#LOG #SPAM_SHIELD\n"
                        f"**User**: {userlink}\n"
                        f"**Banned On**: {chat_link}\n"
                        f"**ID**: {user.id}\n"
                        f"**Reason**: {reason}"
                    ),
                    disable_web_page_preview=True,
                ),
            )
        except BadRequest as err:
            self.log.warning(
                f"Exception raise on {user.id} ({reason} | {banner}) in {chat.id}:{err}",
                exc_info=err,
            )

        return True

    @command.filters(filters.admin_only)
    async def cmd_spamshield(self, ctx: command.Context, enable: Optional[bool] = None) -> str:
        """Set SpamShield setting"""
        chat = ctx.chat
        if not ctx.input:
            return await self.text(chat.id, "spamshield-view", await self.is_active(chat.id))

        if enable is None:
            return await self.text(chat.id, "err-invalid-option")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "spamshield-set", "on" if enable else "off"),
            self.setting(chat.id, enable),
        )
        return ret



"""Bot Greetings"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from html import escape
from typing import Any, ClassVar, MutableMapping, Optional, Tuple

from pyrogram.client import Client
from pyrogram.enums.parse_mode import ParseMode
from pyrogram.errors import ChannelPrivate, ChatWriteForbidden, MessageDeleteForbidden
from pyrogram.types import Chat, Message, User
from pyrogram.types.messages_and_media.message import Str

from anjani import command, filters, plugin, util


class Greeting(plugin.Plugin):
    name: ClassVar[str] = "Greetings"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("WELCOME")

    async def on_chat_action(self, message: Message) -> None:
        chat = message.chat
        reply_to = message.id
        if message.left_chat_member and message.left_chat_member.id == self.bot.uid:
            return

        # Clean service both for left member and new member if active
        if await self.clean_service(chat.id):
            try:
                await message.delete()
            except (MessageDeleteForbidden, ChannelPrivate):
                pass
            reply_to = 0

        if message.new_chat_members:
            return await self._member_join(message, reply_to)

        if message.left_chat_member:
            return await self._member_leave(message, reply_to)

    async def _member_leave(self, message: Message, reply_to: int) -> None:
        chat = message.chat
        if not await self.is_goodbye(chat.id):
            return

        left_member = message.left_chat_member
        text = await self.left_message(chat.id)
        if not text:
            text = await self.text(chat.id, "default-goodbye", noformat=True)

        formatted_text = await self._build_text(text, left_member, chat, self.bot.client)
        try:
            msg = await self.bot.client.send_message(
                chat.id,
                formatted_text,
                reply_to_message_id=reply_to,
            )
        except ChatWriteForbidden:
            return

        previous = await self.previous_goodbye(chat.id, msg.id)
        if previous:
            try:
                await self.bot.client.delete_messages(chat.id, previous)
            except MessageDeleteForbidden:
                pass

    async def _member_join(self, message: Message, reply_to: int) -> None:
        chat = message.chat
        if not await self.is_welcome(chat.id):
            return

        new_members = message.new_chat_members
        for new_member in new_members:
            try:
                if new_member.id == self.bot.uid:
                    await self.bot.client.send_message(
                        chat.id,
                        await self.text(chat.id, "bot-added"),
                        reply_to_message_id=reply_to,
                    )
                else:
                    text, button = await self.welc_message(chat.id)
                    if not text:
                        string = await self.text(chat.id, "default-welcome", noformat=True)
                    else:
                        string = text

                    formatted_text = await self._build_text(
                        string, new_member, chat, self.bot.client
                    )

                    if button:
                        button = util.tg.build_button(button)
                        msg = await self.bot.client.send_message(
                            chat.id,
                            formatted_text,
                            reply_to_message_id=reply_to,
                            reply_markup=button,
                        )
                    else:
                        msg = await self.bot.client.send_message(
                            chat.id,
                            formatted_text,
                            reply_to_message_id=reply_to,
                        )

                    previous = await self.previous_welcome(chat.id, msg.id)
                    if previous:
                        try:
                            await self.bot.client.delete_messages(chat.id, previous)
                        except MessageDeleteForbidden:
                            pass
            except ChatWriteForbidden:
                pass

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        welcome = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: welcome} if welcome else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": data[self.name]}, upsert=True)

    @staticmethod
    async def _build_text(
        text: str, user: User, chat: Chat, client: Optional[Client] = None
    ) -> str:
        first_name = user.first_name or ""  # Ensure first name is not None
        last_name = user.last_name
        full_name = first_name + last_name if last_name else first_name
        try:
            count = await client.get_chat_members_count(chat.id) if client else "N/A"
        except ChannelPrivate:
            count = "N/A"

        return text.format(
            first=escape(first_name),
            last=escape(last_name) if last_name else "",
            fullname=escape(full_name),
            username=f"@{user.username}" if user.username else user.mention,
            mention=user.mention,
            count=count,
            chatname=escape(chat.title),
            id=user.id,
        )

    async def is_welcome(self, chat_id: int) -> bool:
        """Get chat welcome setting"""
        active = await self.db.find_one({"chat_id": chat_id}, {"should_welcome": 1})
        return active.get("should_welcome", True) if active else True

    async def is_goodbye(self, chat_id: int) -> bool:
        """Get chat welcome setting"""
        active = await self.db.find_one({"chat_id": chat_id}, {"should_goodbye": 1})
        return active.get("should_goodbye", True) if active else True

    async def welc_message(
        self, chat_id: int
    ) -> Tuple[Optional[str], Optional[Tuple[Tuple[str, str, bool]]]]:
        """Get chat welcome string"""
        message = await self.db.find_one({"chat_id": chat_id}, {"custom_welcome": 1, "button": 1})
        if message:
            return message.get("custom_welcome"), message.get("button")

        return await self.text(chat_id, "default-welcome", noformat=True), None

    async def left_message(self, chat_id: int) -> str:
        message = await self.db.find_one({"chat_id": chat_id}, {"custom_goodbye": 1})
        return (
            message.get(
                "custom_goodbye", await self.text(chat_id, "default-goodbye", noformat=True)
            )
            if message
            else await self.text(chat_id, "default-goodbye", noformat=True)
        )

    async def clean_service(self, chat_id: int) -> bool:
        """Fetch clean service setting"""
        clean = await self.db.find_one({"chat_id": chat_id}, {"clean_service": 1})
        if clean:
            return clean.get("clean_service", True)

        return False  # Defaults off

    async def set_custom_welcome(self, chat_id: int, text: Str) -> None:
        """Set custom welcome"""
        msg, button = util.tg.parse_button(text.markdown)
        await self.db.update_one(
            {"chat_id": chat_id},
            {"$set": {"custom_welcome": msg, "button": button}},
            upsert=True,
        )

    async def set_custom_goodbye(self, chat_id: int, text: str) -> None:
        """Set custom goodbye"""
        await self.db.update_one({"chat_id": chat_id}, {"$set": {"custom_goodbye": text}})

    async def del_custom_welcome(self, chat_id: int) -> None:
        """Delete custom welcome message"""
        await self.db.update_one(
            {"chat_id": chat_id}, {"$unset": {"custom_welcome": "", "button": ""}}
        )

    async def del_custom_goodbye(self, chat_id: int) -> None:
        """Delete custom goodbye message"""
        await self.db.update_one({"chat_id": chat_id}, {"$unset": {"custom_goodbye": ""}})

    async def greeting_setting(self, chat_id: int, key: str, value: bool) -> None:
        """Turn on/off greetings in chats"""
        if not value:
            await self.db.update_one({"chat_id": chat_id}, {"$set": {key: False}}, upsert=True)
        else:
            await self.db.update_one({"chat_id": chat_id}, {"$unset": {key: ""}}, upsert=True)

    async def previous_welcome(self, chat_id: int, msg_id: int) -> Optional[int]:
        """Save latest welcome msg_id and return previous msg_id"""
        data = await self.db.find_one_and_update(
            {"chat_id": chat_id}, {"$set": {"prev_welc": msg_id}}, upsert=True
        )
        return data.get("prev_welc", None) if data else None

    async def previous_goodbye(self, chat_id: int, msg_id: int) -> Optional[int]:
        data = await self.db.find_one_and_update(
            {"chat_id": chat_id}, {"$set": {"prev_gdby": msg_id}}, upsert=True
        )
        return data.get("prev_gdby", None) if data else None

    @command.filters(filters.admin_only)
    async def cmd_setwelcome(self, ctx: command.Context) -> str:
        """Set chat welcome message"""
        chat = ctx.chat
        if ctx.input:
            welc_text = Str(ctx.input).init(ctx.msg.entities[1:])
        elif ctx.msg.reply_to_message:
            welc_text = ctx.msg.reply_to_message.text or ctx.msg.reply_to_message.caption
        else:
            return await self.text(chat.id, "greetings-no-input")

        try:  # Try to build a text first to check message validity
            await self._build_text(welc_text or "", ctx.author, chat, self.bot.client)
        except (KeyError, ValueError) as e:
            return await self.text(chat.id, "err-msg-format-parsing", err=e)

        ret, _ = await asyncio.gather(
            self.text(chat.id, "cust-welcome-set"), self.set_custom_welcome(chat.id, welc_text)
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_setgoodbye(self, ctx: command.Context) -> str:
        """Set chat goodbye message"""
        chat = ctx.chat
        if ctx.input:
            gby_text = Str(ctx.input).init(ctx.msg.entities[1:])
        elif ctx.msg.reply_to_message:
            gby_text = ctx.msg.reply_to_message.text or ctx.msg.reply_to_message.caption
        else:
            return await self.text(chat.id, "greetings-no-input")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "cust-goodbye-set"), self.set_custom_goodbye(chat.id, gby_text)
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_resetwelcome(self, ctx: command.Context) -> str:
        """Reset saved welcome message"""
        chat = ctx.chat

        ret, _ = await asyncio.gather(
            self.text(chat.id, "reset-welcome"), self.del_custom_welcome(chat.id)
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_resetgoodbye(self, ctx: command.Context) -> str:
        """Reset saved welcome message"""
        chat = ctx.chat

        ret, _ = await asyncio.gather(
            self.text(chat.id, "reset-goodbye"), self.del_custom_goodbye(chat.id)
        )
        return ret

    @command.filters(filters.admin_only)
    async def cmd_welcome(self, ctx: command.Context) -> Optional[str]:
        """View current welcome message"""
        chat = ctx.chat
        param = ctx.input.lower()
        noformat = param == "noformat"

        enabled = None
        if param in {"yes", "on", "1"}:
            enabled = True
        elif param in {"no", "off", "0"}:
            enabled = False
        elif param and not noformat:
            return await self.text(chat.id, "err-invalid-option")

        if enabled is not None:
            ret, _ = await asyncio.gather(
                self.text(chat.id, "welcome-set", "on" if enabled else "off"),
                self.greeting_setting(chat.id, "should_welcome", enabled),
            )
            return ret

        setting, (text, button), clean_service = await asyncio.gather(
            self.is_welcome(chat.id), self.welc_message(chat.id), self.clean_service(chat.id)
        )

        if text is None:
            text = ""
        else:
            text += "\n\n"

        if noformat:
            parse_mode = ParseMode.DISABLED
            if button:
                text += util.tg.revert_button(button)
            button = None
        else:
            parse_mode = ParseMode.MARKDOWN
            if button:
                button = util.tg.build_button(button)

        await ctx.respond(await self.text(chat.id, "view-welcome", setting, clean_service))
        await ctx.respond(
            text
            if text
            else (
                "Empty, custom welcome message haven't set yet."
                if not setting
                else "Default:\n\n" + await self.text(chat.id, "default-welcome", noformat=True)
            ),
            mode="reply",
            reply_markup=button,
            parse_mode=parse_mode,
            disable_web_page_preview=True,
        )
        return None

    @command.filters(filters.admin_only)
    async def cmd_goodbye(self, ctx: command.Context) -> Optional[str]:
        """View current goodbye message"""
        chat = ctx.chat
        param = ctx.input.lower()
        noformat = param == "noformat"

        enabled = None
        if param in {"yes", "on", "1"}:
            enabled = True
        elif param in {"no", "off", "0"}:
            enabled = False
        elif param and not noformat:
            return await self.text(chat.id, "err-invalid-option")

        if enabled is not None:
            ret, _ = await asyncio.gather(
                self.text(chat.id, "goodbye-set", "on" if enabled else "off"),
                self.greeting_setting(chat.id, "should_goodbye", enabled),
            )
            return ret

        setting, text, clean_service = await asyncio.gather(
            self.is_goodbye(chat.id), self.left_message(chat.id), self.clean_service(chat.id)
        )

        if noformat:
            parse_mode = ParseMode.DISABLED
        else:
            parse_mode = ParseMode.MARKDOWN

        await ctx.respond(await self.text(chat.id, "view-goodbye", setting, clean_service))
        await ctx.respond(
            text,
            mode="reply",
            parse_mode=parse_mode,
            disable_web_page_preview=True,
        )
        return None

    @command.filters(filters.admin_only)
    async def cmd_cleanservice(self, ctx: command.Context, active: Optional[bool] = None) -> str:
        """Clean service message on new members"""
        chat = ctx.chat

        if active is None:
            return await self.text(chat.id, "err-invalid-option")

        ret, _ = await asyncio.gather(
            self.text(chat.id, "clean-serv-set", "on" if active else "off"),
            self.greeting_setting(chat.id, "clean_service", active),
        )
        return ret



"""Anjani filters"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, Callable, Coroutine, Optional

from pyrogram.client import Client
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_type import ChatType
from pyrogram.filters import (  # skipcq: PY-W2000
    Filter,
    animation,
    audio,
    bot,
    caption,
    channel,
    channel_chat_created,
    chat,
    command,
    contact,
    delete_chat_photo,
    dice,
    document,
    forwarded,
    from_scheduled,
    game,
    game_high_score,
    group,
    group_chat_created,
    incoming,
    inline_keyboard,
    left_chat_member,
    linked_channel,
    location,
    me,
    media,
    media_group,
    mentioned,
    migrate_from_chat_id,
    migrate_to_chat_id,
    new_chat_members,
    new_chat_photo,
    new_chat_title,
    outgoing,
    photo,
    pinned_message,
    poll,
    private,
    regex,
    reply,
    reply_keyboard,
    scheduled,
    service,
    sticker,
    supergroup_chat_created,
    text,
    user,
    venue,
    via_bot,
    video,
    video_chat_ended,
    video_chat_members_invited,
    video_chat_started,
    video_note,
    web_page,
)
from pyrogram.types import ChatMember, Message

from anjani.util.tg import fetch_permissions, get_text, reply_and_delete
from anjani.util.types import CustomFilter

if TYPE_CHECKING:
    from anjani.core import Anjani


FilterFunc = Callable[[CustomFilter, Client, Message], Coroutine[Any, Any, bool]]
__all__ = [
    "admin_only",
    "animation",
    "audio",
    "bot",
    "can_change_info",
    "can_delete",
    "can_invite",
    "can_pin",
    "can_promote",
    "can_restrict",
    "caption",
    "channel",
    "channel_chat_created",
    "chat",
    "command",
    "contact",
    "delete_chat_photo",
    "dev_only",
    "dice",
    "document",
    "forwarded",
    "from_scheduled",
    "game",
    "game_high_score",
    "group",
    "group_chat_created",
    "incoming",
    "inline_keyboard",
    "left_chat_member",
    "linked_channel",
    "location",
    "me",
    "media",
    "media_group",
    "mentioned",
    "migrate_from_chat_id",
    "migrate_to_chat_id",
    "new_chat_members",
    "new_chat_photo",
    "new_chat_title",
    "outgoing",
    "owner_only",
    "photo",
    "pinned_message",
    "poll",
    "private",
    "regex",
    "reply",
    "reply_keyboard",
    "scheduled",
    "service",
    "staff_only",
    "sticker",
    "supergroup_chat_created",
    "text",
    "user",
    "web_page",
    "venue",
    "via_bot",
    "video",
    "video_note",
    "video",
    "video_chat_ended",
    "video_chat_members_invited",
    "video_chat_started",
]


def create(func: FilterFunc, name: Optional[str] = None, **kwargs: Any) -> CustomFilter:
    return type(
        name or func.__name__ or "CustomAnjaniFilter", (CustomFilter,), {"__call__": func, **kwargs}
    )()


# { permission
def _create_filter_permission(name: str, *, include_bot: bool = True) -> Filter:
    async def func(flt: CustomFilter, client: Client, message: Message) -> bool:
        target, priv = message.from_user, message.chat and message.chat.type == ChatType.PRIVATE
        if priv or not target or not message.chat:
            return False

        bot_perm, member_perm = await fetch_permissions(client, message.chat.id, target.id)
        if not (bot_perm and member_perm) or not (bot_perm.privileges and member_perm.privileges):
            return False

        try:
            if getattr(bot_perm.privileges, name) and getattr(member_perm.privileges, name):
                return True
        except AttributeError:
            flt.anjani.log.error(f"{name} is not a valid permission")
            return False

        flt.anjani.loop.create_task(
            reply_and_delete(
                message, await get_text(flt.anjani, message.chat.id, "err-perm", name), 5
            )
        )
        return False

    return create(func, name, include_bot=include_bot)


can_change_info = _create_filter_permission("can_change_info")
can_delete = _create_filter_permission("can_delete_messages")
can_invite = _create_filter_permission("can_invite_users")
can_pin = _create_filter_permission("can_pin_messages")
can_promote = _create_filter_permission("can_promote_members")
can_restrict = _create_filter_permission("can_restrict_members")
# }


# { staff_only
def _staff_only(include_bot: bool = True, *, rank: Optional[str] = None) -> CustomFilter:
    async def func(flt: CustomFilter, _: Client, message: Message) -> bool:
        target = message.from_user
        if not target:  # Sanity check for anonymous admin
            return False

        if rank is None:
            return target.id in flt.anjani.staff

        if rank == "dev":
            return target.id in flt.anjani.devs

        flt.anjani.log.error(f"Invalid rank: {rank}")
        return False

    return create(func, "staff_only", include_bot=include_bot)


staff_only = _staff_only()
dev_only = _staff_only(rank="dev")
# }


# { owner_only
def _owner_only(include_bot: bool = True) -> CustomFilter:
    async def func(flt: CustomFilter, _: Client, message: Message) -> bool:
        target = message.from_user
        if not target:  # Sanity check for anonymous admin
            return False

        return target.id == flt.anjani.owner

    return create(func, "owner_only", include_bot=include_bot)


owner_only = _owner_only()
# }


# { admin_only
async def _send_error(robot: "Anjani", chat_id: int, message: Message, string_key: str) -> None:
    robot.loop.create_task(reply_and_delete(message, await get_text(robot, chat_id, string_key), 5))


def is_admin(target: ChatMember) -> bool:
    return target.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}


def _admin_only(include_bot: bool = True, send_error: bool = True) -> CustomFilter:
    async def func(flt: CustomFilter, client: Client, message: Message) -> bool:
        target, priv = message.from_user, message.chat and message.chat.type == ChatType.PRIVATE
        if priv or not message.chat:
            return False

        if not target:
            if message.sender_chat:
                if message.sender_chat.id == message.chat.id:  # Anonymous Admin
                    return True

                curr_chat: Any = await client.get_chat(message.chat.id)
                if (
                    curr_chat.linked_chat
                    and message.sender_chat.id == curr_chat.linked_chat.id
                    and not message.forward_from_chat
                ):  # Linked Channel Owner
                    return True

            return False

        bot_perm, member_perm = await fetch_permissions(client, message.chat.id, target.id)
        if not bot_perm:
            if send_error:
                await _send_error(flt.anjani, message.chat.id, message, "err-im-not-admin")

            return False

        if not member_perm:
            if send_error:
                await _send_error(flt.anjani, message.chat.id, message, "err-not-admin")

            return False

        user_admin = is_admin(member_perm)
        bot_admin = is_admin(bot_perm)
        if bot_admin and user_admin:
            return True

        if send_error:
            if not bot_admin and user_admin:
                # Bot is not admin, but user is
                await _send_error(flt.anjani, message.chat.id, message, "err-im-not-admin")
            elif bot_admin and not user_admin:
                # Bot is admin, but user is not
                await _send_error(flt.anjani, message.chat.id, message, "err-not-admin")
            else:
                await _send_error(flt.anjani, message.chat.id, message, "err-perm")

        return False

    return create(func, "admin_only", include_bot=include_bot)


admin_only = _admin_only()
admin_only_no_report = _admin_only(send_error=False)
"""Set filter to admin only but without sending error message"""
# }



""" Filters plugin. """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import re
from typing import (
    Any,
    Callable,
    ClassVar,
    Coroutine,
    MutableMapping,
    Optional,
    Set,
    Tuple,
)

from pyrogram.errors import MediaEmpty, MessageEmpty
from pyrogram.types import Message

from anjani import command, filters, listener, plugin, util
from anjani.util.tg import Types, build_button, get_message_info


class Filters(plugin.Plugin):
    name: ClassVar[str] = "Filters"
    helpable: ClassVar[bool] = True

    db: util.db.AsyncCollection
    trigger: MutableMapping[int, Set[str]] = {}
    SEND: MutableMapping[int, Callable[..., Coroutine[Any, Any, Optional[Message]]]]

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("FILTERS")
        self.SEND = {
            Types.TEXT.value: self.bot.client.send_message,
            Types.BUTTON_TEXT.value: self.bot.client.send_message,
            Types.DOCUMENT.value: self.bot.client.send_document,
            Types.PHOTO.value: self.bot.client.send_photo,
            Types.VIDEO.value: self.bot.client.send_video,
            Types.STICKER.value: self.bot.client.send_sticker,
            Types.AUDIO.value: self.bot.client.send_audio,
            Types.VOICE.value: self.bot.client.send_voice,
            Types.VIDEO_NOTE.value: self.bot.client.send_video_note,
            Types.ANIMATION.value: self.bot.client.send_animation,
        }

    async def on_start(self, _: int) -> None:
        async for chat in self.db.find({}):
            self.trigger[chat["chat_id"]] = set(chat["trigger"].keys())

    async def on_plugin_backup(self, chat_id: int) -> MutableMapping[str, Any]:
        data = await self.db.find_one({"chat_id": chat_id}, {"_id": False})
        return {self.name: data} if data else {}

    async def on_plugin_restore(self, chat_id: int, data: MutableMapping[str, Any]) -> None:
        await self.db.update_one({"chat_id": chat_id}, {"$set": {data[self.name]}}, upsert=True)

    async def on_chat_migrate(self, message: Message) -> None:
        new_chat = message.chat.id
        old_chat = message.migrate_from_chat_id

        await self.db.update_one(
            {"chat_id": old_chat},
            {"$set": {"chat_id": new_chat}},
        )

    @listener.priority(95)
    async def on_message(self, message: Message) -> None:
        if message.outgoing:
            return

        chat = message.chat
        text = message.text or message.caption

        if not (text or chat):
            return

        chat_trigger = self.trigger.get(chat.id, [])
        if not chat_trigger:
            return

        await self.reply_filter(message, set(chat_trigger), text)

    async def reply_filter(self, message: Message, trigger: Set[str], text: str):
        if not text or text.startswith("/filter") or text.startswith("/stop"):
            return  # Igonore when command triggered
        for i in trigger:
            pattern = r"( |^|[^\w])" + re.escape(i) + r"( |$|[^\w])"
            if re.search(pattern, text, flags=re.IGNORECASE):
                filt = await self.get_filter(message.chat.id, i)
                if not filt:
                    return

                # This checks data for old filters schema
                # TODO: deprecate old schema on v3
                if isinstance(filt, str):
                    await message.reply_text(filt)
                else:
                    reply_to = (
                        message.reply_to_message.id if message.reply_to_message else message.id
                    )
                    types: int = filt["type"]
                    button = filt.get("button", None)
                    if button:
                        keyb = build_button(button)
                    else:
                        keyb = button

                    try:
                        if types in {Types.TEXT, Types.BUTTON_TEXT}:
                            await self.SEND[types](
                                message.chat.id,
                                filt["text"],
                                reply_to_message_id=reply_to,
                                reply_markup=keyb,
                            )
                        elif types in {Types.STICKER, Types.ANIMATION}:
                            await self.SEND[types](
                                message.chat.id,
                                filt["content"],
                                reply_to_message_id=reply_to,
                            )
                        else:
                            await self.SEND[types](
                                message.chat.id,
                                filt["content"],
                                caption=filt["text"],
                                reply_to_message_id=reply_to,
                                reply_markup=keyb,
                            )
                    except MediaEmpty:
                        await self.bot.client.send_message(
                            message.chat.id, await self.text(message.chat.id, "notes-expired")
                        )
                    except MessageEmpty:
                        self.log.warning(
                            "Filter message empty on %s with data %s", message.chat.id, filt
                        )
                break

    async def get_filter(self, chat_id: int, keyword: str) -> Optional[str]:
        data = await self.db.find_one(
            {"chat_id": chat_id, f"trigger.{keyword}": {"$exists": True}},
            {f"trigger.{keyword}": 1},
        )
        return data["trigger"][keyword] if data else None

    async def del_filter(self, chat_id: int, keyword: str) -> Tuple[bool, str]:
        filt = self.trigger.get(chat_id)
        if not filt:
            return False, await self.text(chat_id, "filters-chat-nofilter")
        if keyword not in filt:
            return False, await self.text(chat_id, "filters-chat-nokeyword", keyword)

        await self.db.update_one(
            {"chat_id": chat_id},
            {"$unset": {f"trigger.{keyword}": ""}},
        )
        self.trigger[chat_id].remove(keyword)
        return True, ""

    @command.filters(filters.admin_only)
    async def cmd_filter(self, ctx: command.Context) -> str:
        chat = ctx.chat
        if (
            len(ctx.args) < 2
            and not ctx.message.reply_to_message
            or ctx.msg.reply_to_message
            and len(ctx.args) < 1
        ):
            return await self.text(chat.id, "filter-help")

        trigger = ctx.args[0]

        if "." in trigger or "$" in trigger:
            return await self.text(chat.id, "err-illegal-trigger")

        text, types, content, buttons = get_message_info(ctx.msg)
        _, ret = await asyncio.gather(
            self.db.update_one(
                {"chat_id": chat.id},
                {
                    "$set": {
                        f"trigger.{trigger}": {
                            "text": text,
                            "type": types,
                            "content": content,
                            "buttons": buttons,
                        }
                    }
                },
                upsert=True,
            ),
            self.text(chat.id, "filters-added", trigger),
        )

        if self.trigger.get(chat.id):
            self.trigger[chat.id].add(trigger)
        else:
            self.trigger[chat.id] = {trigger}

        return ret

    @command.filters(filters.admin_only)
    async def cmd_stop(self, ctx: command.Context, trigger: str) -> str:
        if not trigger:
            return await self.text(ctx.chat.id, "filter-stop-help")

        deleted, out = await self.del_filter(ctx.chat.id, trigger)
        if not deleted:
            return out

        return await self.text(ctx.chat.id, "filters-removed", trigger)

    @command.filters(filters.admin_only, aliases=["rmallfilters"])
    async def cmd_rmallfilter(self, ctx: command.Context) -> str:
        chat_id = ctx.chat.id
        triggers = self.trigger.pop(chat_id, None)
        if not triggers:
            return await self.text(chat_id, "filters-chat-nofilter")
        await self.db.delete_one({"chat_id": chat_id})
        return await self.text(chat_id, "filters-rmall", len(triggers))

    @command.filters(filters.admin_only)
    async def cmd_filters(self, ctx: command.Context) -> str:
        data = self.trigger.get(ctx.chat.id)
        if not data:
            return await self.text(ctx.chat.id, "filters-chat-nofilter")

        output = await self.text(ctx.chat.id, "filters-list", ctx.chat.title)
        output += "\n".join([f"× `{i}`" for i in sorted(data)])
        return output



""" Debugging purpose """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
import io
import os
import re
import sys
import traceback
from datetime import datetime
from html import escape
from typing import Any, ClassVar, Optional, Tuple

import pyrogram
from meval import meval
from pyrogram.enums.chat_action import ChatAction

from anjani import command, filters, plugin, util


class Debug(plugin.Plugin):
    name: ClassVar[str] = "Debug"

    async def cmd_ping(self, ctx: command.Context) -> str:
        start = datetime.now()
        await ctx.respond("Calculating response time...")
        end = datetime.now()
        latency = (end - start).microseconds / 1000

        return f"Latency: {latency} ms"

    @command.filters(filters.dev_only)
    async def cmd_eval(self, ctx: command.Context) -> Optional[str]:
        code = ctx.input
        if not code:
            return "Give me code to evaluate."

        out_buf = io.StringIO()

        async def _eval() -> Tuple[str, Optional[str]]:
            # Message sending helper for convenience
            async def send(*args: Any, **kwargs: Any) -> pyrogram.types.Message:
                return await ctx.msg.reply(*args, **kwargs)

            # Print wrapper to capture output
            # We don't override sys.stdout to avoid interfering with other output
            def _print(*args: Any, **kwargs: Any) -> None:
                if "file" not in kwargs:
                    kwargs["file"] = out_buf

                return print(*args, **kwargs)

            eval_vars = {
                # Contextual info
                "self": self,
                "ctx": ctx,
                "bot": self.bot,
                "loop": self.bot.loop,
                "client": self.bot.client,
                "commands": self.bot.commands,
                "listeners": self.bot.listeners,
                "plugins": self.bot.plugins,
                "stdout": out_buf,
                # Convenience aliases
                "anjani": self.bot,
                "chat": ctx.chat,
                "context": ctx,
                "msg": ctx.msg,
                "message": ctx.msg,
                "db": self.bot.db,
                # Helper functions
                "send": send,
                "print": _print,
                # Built-in modules
                "inspect": inspect,
                "os": os,
                "re": re,
                "sys": sys,
                "traceback": traceback,
                # Third-party modules
                "pyrogram": pyrogram,
                # Custom modules
                "command": command,
                "plugin": plugin,
                "util": util,
            }

            try:
                return "", await meval(code, globals(), **eval_vars)
            except Exception as e:  # skipcq: PYL-W0703
                # Find first traceback frame involving the snippet
                first_snip_idx = -1
                tb = traceback.extract_tb(e.__traceback__)
                for i, frame in enumerate(tb):
                    if frame.filename == "<string>" or frame.filename.endswith("ast.py"):
                        first_snip_idx = i
                        break

                # Re-raise exception if it wasn't caused by the snippet
                if first_snip_idx == -1:
                    raise e

                # Return formatted stripped traceback
                stripped_tb = tb[first_snip_idx:]
                formatted_tb = util.error.format_exception(e, tb=stripped_tb)
                return "⚠️ Error executing snippet\n\n", formatted_tb

        before = util.time.usec()
        prefix, result = await _eval()
        after = util.time.usec()

        # Always write result if no output has been collected thus far
        if not out_buf.getvalue() or result is not None:
            print(result, file=out_buf)

        el_us = after - before
        el_str = util.time.format_duration_us(el_us)

        out = out_buf.getvalue()
        # Strip only ONE final newline to compensate for our message formatting
        if out.endswith("\n"):
            out = out[:-1]

        if len(out) > 4096:
            async with ctx.action(ChatAction.UPLOAD_DOCUMENT):
                with io.BytesIO(str.encode(out)) as out_file:
                    out_file.name = "eval.text"
                    await ctx.msg.reply_document(
                        document=out_file, caption=code, disable_notification=True
                    )

                return None

        await ctx.respond(
            f"""{prefix}<b>In:</b>
<pre language="python">{escape(code)}</pre>

<b>Out:</b>
<pre language="python">{escape(out)}</pre>

Time: {el_str}""",
            parse_mode=pyrogram.enums.parse_mode.ParseMode.HTML,
        )



"""Anjani database session"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import asynccontextmanager
from time import monotonic as monotonic_time
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Callable,
    Coroutine,
    Mapping,
    Optional,
)

from bson.timestamp import Timestamp
from pymongo.client_session import ClientSession, SessionOptions
from pymongo.read_concern import ReadConcern
from pymongo.write_concern import WriteConcern

from anjani import util

from .base import AsyncBase
from .errors import OperationFailure, PyMongoError
from .typings import ReadPreferences, Results

if TYPE_CHECKING:
    from .client import AsyncClient


class AsyncClientSession(AsyncBase):
    """AsyncIO :obj:`~ClientSession`

    *DEPRECATED* methods are removed in this class.
    """

    _client: "AsyncClient"

    dispatch: ClientSession

    def __init__(self, client: "AsyncClient", dispatch: ClientSession) -> None:
        self._client = client

        # Propagate initialization to base
        super().__init__(dispatch)

    async def __aenter__(self) -> "AsyncClientSession":
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await util.run_sync(self.dispatch.__exit__, exc_type, exc_val, exc_tb)

    def __enter__(self) -> None:
        raise RuntimeError("Use 'async with' not just 'with'")

    async def abort_transaction(self) -> None:
        return await util.run_sync(self.dispatch.abort_transaction)

    async def commit_transaction(self) -> None:
        return await util.run_sync(self.dispatch.commit_transaction)

    async def end_session(self) -> None:
        return await util.run_sync(self.dispatch.end_session)

    @asynccontextmanager
    async def start_transaction(
        self,
        *,
        read_concern: Optional[ReadConcern] = None,
        write_concern: Optional[WriteConcern] = None,
        read_preference: Optional[ReadPreferences] = None,
        max_commit_time_ms: Optional[int] = None,
    ) -> AsyncGenerator["AsyncClientSession", None]:
        await util.run_sync(
            self.dispatch.start_transaction,
            read_concern=read_concern,
            write_concern=write_concern,
            read_preference=read_preference,
            max_commit_time_ms=max_commit_time_ms,
        )
        try:
            yield self
        except Exception:  # skipcq: PYL-W0703
            if self.in_transaction:
                await self.abort_transaction()
        else:
            if self.in_transaction:
                await self.commit_transaction()

    async def with_transaction(
        self,
        callback: Callable[["AsyncClientSession"], Coroutine[Any, Any, Results]],
        *,
        read_concern: Optional[ReadConcern] = None,
        write_concern: Optional[WriteConcern] = None,
        read_preference: Optional[ReadPreferences] = None,
        max_commit_time_ms: Optional[int] = None,
    ) -> Results:
        # 99% Of this code from motor's lib

        def _within_time_limit(s: float) -> bool:
            return monotonic_time() - s < 120

        def _max_time_expired_error(exc: PyMongoError) -> bool:
            return isinstance(exc, OperationFailure) and exc.code == 50

        start_time = monotonic_time()
        while True:
            async with self.start_transaction(
                read_concern=read_concern,
                write_concern=write_concern,
                read_preference=read_preference,
                max_commit_time_ms=max_commit_time_ms,
            ):
                try:
                    ret = await callback(self)
                except Exception as exc:
                    if self.in_transaction:
                        await self.abort_transaction()
                    if (
                        isinstance(exc, PyMongoError)
                        and exc.has_error_label("TransientTransactionError")
                        and _within_time_limit(start_time)
                    ):
                        # Retry the entire transaction.
                        continue
                    raise

            if not self.in_transaction:
                # Assume callback intentionally ended the transaction.
                return ret

            while True:
                try:
                    await self.commit_transaction()
                except PyMongoError as exc:
                    if (
                        exc.has_error_label("UnknownTransactionCommitResult")
                        and _within_time_limit(start_time)
                        and not _max_time_expired_error(exc)
                    ):
                        # Retry the commit.
                        continue

                    if exc.has_error_label("TransientTransactionError") and _within_time_limit(
                        start_time
                    ):
                        # Retry the entire transaction.
                        break
                    raise

                # Commit succeeded.
                return ret

    def advance_cluster_time(self, cluster_time: Mapping[str, Any]) -> None:
        self.dispatch.advance_cluster_time(cluster_time=cluster_time)

    def advance_operation_time(self, operation_time: Timestamp) -> None:
        self.dispatch.advance_operation_time(operation_time=operation_time)

    @property
    def client(self) -> "AsyncClient":
        return self._client

    @property
    def cluster_time(self) -> Optional[Mapping[str, Any]]:
        return self.dispatch.cluster_time

    @property
    def has_ended(self) -> bool:
        return self.dispatch.has_ended

    @property
    def in_transaction(self) -> bool:
        return self.dispatch.in_transaction

    @property
    def operation_time(self) -> Optional[Timestamp]:
        return self.dispatch.operation_time

    @property
    def options(self) -> SessionOptions:
        return self.dispatch.options

    @property
    def session_id(self) -> Mapping[str, Any]:
        return self.dispatch.session_id



""" Canonical plugin for @dAnjani_bot """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
from typing import Any, ClassVar, MutableMapping

from pymongo.errors import PyMongoError
from pyrogram.enums.chat_member_status import ChatMemberStatus
from pyrogram.enums.chat_members_filter import ChatMembersFilter
from pyrogram.enums.message_media_type import MessageMediaType
from pyrogram.types import (
    ChatMemberUpdated,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    LoginUrl,
    Message,
)

try:
    import userbotindo

    _run_canonical = True
    del userbotindo
except ImportError:
    _run_canonical = False


from anjani import command, filters, listener, plugin, util


class Canonical(plugin.Plugin):
    """Helper Plugin
    This plugin is only available for @dAnjani_bot
    to comunicate with https://userbotindo.com
    """

    name: ClassVar[str] = "Canonical"
    disabled: ClassVar[bool] = not _run_canonical

    # Private
    __task: asyncio.Task[None]
    _mt: MutableMapping[MessageMediaType, str] = {
        MessageMediaType.STICKER: "s",
        MessageMediaType.PHOTO: "p",
        MessageMediaType.DOCUMENT: "f",
        MessageMediaType.VIDEO: "v",
    }

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("TEST")
        self.db_analytics = self.bot.db.get_collection("ANALYTICS")
        self.chats_db = self.bot.db.get_collection("CHATS")

    async def on_start(self, _: int) -> None:
        self.log.debug("Starting watch streams")
        self.__task = self.bot.loop.create_task(self.watch_streams())

    async def on_stop(self) -> None:
        self.log.debug("Stopping watch streams")
        self.__task.cancel()

    def get_type(self, message: Message) -> str:
        return self._mt.get(message.media, "t") if message.media else "t"

    async def save_message_type(self, message: Message) -> None:
        today = util.time.sec()
        timestamp = today - (today % 86400)  # truncate to day
        await self.db_analytics.update_one(
            {"key": 2},
            {"$inc": {f"data.{str(timestamp)}.{self.get_type(message)}": 1}},
            upsert=True,
        )

    @command.filters(filters.admin_only & filters.group)
    async def cmd_r(self, ctx: command.Context) -> None:
        """Refresh chat data"""
        admins = []
        async for member in self.bot.client.get_chat_members(
            ctx.chat.id, filter=ChatMembersFilter.ADMINISTRATORS
        ):  # type: ignore
            if not member.user.is_bot and member.privileges.can_manage_chat:
                admins.append(member.user.id)

        await self.chats_db.update_one(
            {"chat_id": ctx.chat.id}, {"$set": {"admins": admins}}, upsert=True
        )
        await ctx.respond("Done", delete_after=5)

    @listener.priority(65)
    async def on_message(self, message: Message) -> None:
        if message.outgoing:
            return

        # Analytics
        self.bot.loop.create_task(self.save_message_type(message))

    async def on_chat_action(self, message: Message) -> None:
        """Delete admins data from chats"""
        if message.new_chat_members:
            return

        chat = message.chat
        user = message.left_chat_member
        if user.id == self.bot.uid:
            await asyncio.gather(
                self.chats_db.update_one({"chat_id": chat.id}, {"$set": {"admins": []}}),
            )

    async def on_chat_member_update(self, update: ChatMemberUpdated) -> None:
        old_data = update.old_chat_member
        new_data = update.new_chat_member

        if not old_data or not new_data:  # Rare case
            return

        if old_data.status == new_data.status:
            return

        if new_data.status not in {
            ChatMemberStatus.ADMINISTRATOR,
            ChatMemberStatus.OWNER,
        }:
            await self.chats_db.update_one(
                {"chat_id": update.chat.id}, {"$pull": {"admins": new_data.user.id}}
            )
        elif (
            new_data.status in {ChatMemberStatus.ADMINISTRATOR, ChatMemberStatus.OWNER}
            and new_data.privileges
            and new_data.privileges.can_manage_chat
        ):  # type: ignore
            await self.chats_db.update_one(
                {"chat_id": update.chat.id},
                {"$addToSet": {"admins": new_data.user.id}},
                upsert=True,
            )

    async def watch_streams(self) -> None:
        try:
            async with self.db.watch([{"$match": {"operationType": "insert"}}]) as stream:
                async for change in stream:
                    await self.dispatch_change(change["fullDocument"])
        except PyMongoError as e:
            self.log.error("Error", exc_info=e)

    async def dispatch_change(self, doc: MutableMapping[str, Any]) -> None:
        chat_id = int(doc["_id"])
        message = doc["message"]
        pin = doc.get("pin", False)
        disable_preview = doc.get("disable_preview", False)

        try:
            msg = await self.bot.client.send_message(
                chat_id=chat_id,
                text=message,
                disable_web_page_preview=disable_preview,
            )
            if pin:
                await self.bot.client.pin_chat_message(msg.chat.id, msg.id)
        except Exception as e:  # skipcq: PYL-W0703
            self.log.error(f"Error sending message to {chat_id}", exc_info=e)
        finally:
            await self.db.delete_one({"_id": chat_id})

    @command.filters(filters.private)
    async def cmd_login(self, ctx: command.Context):
        """Login to https://userbotindo.com"""
        if not self.bot.config["login_url"]:
            return

        await ctx.respond(
            "Click this button to login to Anjani Dashboard",
            reply_markup=InlineKeyboardMarkup(
                [
                    [
                        InlineKeyboardButton(
                            "Login",
                            login_url=LoginUrl(
                                url=self.bot.config["login_url"],
                                forward_text="Login to https://userbotindo.com",
                                request_write_access="True",
                            ),
                        )
                    ]
                ]
            ),
        )



""" Main Anjani plugins """
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import re
from hashlib import sha256
from typing import TYPE_CHECKING, Any, ClassVar, List, Optional

from aiopath import AsyncPath
from bson.binary import Binary
from pyrogram.enums.chat_type import ChatType
from pyrogram.enums.parse_mode import ParseMode
from pyrogram.errors import (
    ChannelInvalid,
    ChannelPrivate,
    MessageDeleteForbidden,
    MessageNotModified,
)
from pyrogram.raw.functions.updates.get_state import GetState
from pyrogram.types import (
    CallbackQuery,
    InlineKeyboardButton,
    InlineKeyboardMarkup,
    Message,
)

from anjani import command, filters, listener, plugin, util

if TYPE_CHECKING:
    from .rules import Rules


class Main(plugin.Plugin):
    """Bot main Commands"""

    name: ClassVar[str] = "Main"

    bot_name: str
    db: util.db.AsyncCollection

    async def on_load(self) -> None:
        self.db = self.bot.db.get_collection("SESSION")

    async def on_start(self, _: int) -> None:
        self.bot_name = (
            self.bot.user.first_name + self.bot.user.last_name
            if self.bot.user.last_name
            else self.bot.user.first_name
        )

        restart = await self.db.find_one({"_id": 5})
        if restart is not None:
            rs_time: Optional[int] = restart.get("time")
            rs_chat_id: Optional[int] = restart.get("status_chat_id")
            rs_message_id: Optional[int] = restart.get("status_message_id")

            # Delete data first in case message editing fails
            await self.db.delete_one({"_id": 5})

            # Bail out if we're missing necessary values
            if rs_chat_id is None or rs_message_id is None or rs_time is None:
                return

            duration = util.time.usec() - rs_time
            duration_str = util.time.format_duration_us(duration)
            __, status_msg = await asyncio.gather(
                self.bot.log_stat("downtime", value=duration),
                self.bot.client.get_messages(rs_chat_id, rs_message_id),
            )
            if isinstance(status_msg, List):
                status_msg = status_msg[0]

            self.bot.log.info(f"Bot downtime {duration_str}")
            await self.send_to_log(
                f"Bot downtime {duration_str}.", reply_to_message_id=status_msg.id
            )
            try:
                await status_msg.delete()
            except MessageDeleteForbidden:
                pass
        else:
            await self.send_to_log("Starting system...")

    async def on_stop(self) -> None:
        file = AsyncPath("anjani/anjani.session")
        if not await file.exists():
            return

        data = await self.bot.client.invoke(GetState())
        await self.db.update_one(
            {"_id": sha256(self.bot.config["api_id"].encode()).hexdigest()},
            {
                "$set": {
                    "session": Binary(await file.read_bytes()),
                    "date": data.date,
                    "pts": data.pts,
                    "qts": data.qts,
                    "seq": data.seq,
                }
            },
            upsert=True,
        )

        status_msg = await self.send_to_log("Shutdowning system...")
        self.bot.log.info("Preparing to shutdown...")
        if not status_msg:
            return

        await self.db.update_one(
            {"_id": 5},
            {
                "$set": {
                    "status_chat_id": status_msg.chat.id,
                    "status_message_id": status_msg.id,
                    "time": util.time.usec(),
                }
            },
            upsert=True,
        )

    async def send_to_log(self, text: str, *args: Any, **kwargs: Any) -> Optional[Message]:
        try:
            return await self.bot.client.send_message(
                int(self.bot.config.log_channel), text, *args, **kwargs
            )
        except AttributeError:
            pass

    async def help_builder(self, chat_id: int) -> List[List[InlineKeyboardButton]]:
        """Build the help button"""
        plugins: List[InlineKeyboardButton] = []
        for plug in list(self.bot.plugins.values()):
            if plug.helpable:
                plugins.append(
                    InlineKeyboardButton(
                        await self.text(chat_id, f"{plug.name.lower()}-button"),
                        callback_data=f"help_plugin({plug.name.lower()})",
                    )
                )
        plugins.sort(key=lambda kb: kb.text)

        pairs = [plugins[i * 3 : (i + 1) * 3] for i in range((len(plugins) + 3 - 1) // 3)]
        pairs.append([InlineKeyboardButton("✗ Close", callback_data="help_close")])

        return pairs

    @listener.filters(filters.regex(r"help_(.*)"))
    async def on_callback_query(self, query: CallbackQuery) -> None:
        """Bot helper button"""
        match = query.matches[0].group(1)
        chat = query.message.chat

        if match == "back":
            keyboard = await self.help_builder(chat.id)
            try:
                await query.edit_message_text(
                    await self.text(chat.id, "help-pm", self.bot_name),
                    reply_markup=InlineKeyboardMarkup(keyboard),
                    parse_mode=ParseMode.MARKDOWN,
                )
            except MessageNotModified:
                pass
        elif match == "close":
            try:
                await query.message.delete()
            except MessageDeleteForbidden:
                await query.answer("I can't delete the message")
        elif match:
            plug = re.compile(r"plugin\((\w+)\)").match(match)
            if not plug:
                raise ValueError("Unable to find plugin name")

            text_lang = await self.text(
                chat.id, f"{plug.group(1)}-help", username=self.bot.user.username
            )
            text = (
                f"Here is the help for the **{plug.group(1).capitalize()}** "
                f"plugin:\n\n{text_lang}"
            )
            try:
                await query.edit_message_text(
                    text,
                    reply_markup=InlineKeyboardMarkup(
                        [
                            [
                                InlineKeyboardButton(
                                    await self.text(chat.id, "back-button"),
                                    callback_data="help_back",
                                )
                            ]
                        ]
                    ),
                    parse_mode=ParseMode.MARKDOWN,
                )
            except MessageNotModified:
                pass

    async def cmd_start(self, ctx: command.Context) -> Optional[str]:
        """Bot start command"""
        chat = ctx.chat

        if chat.type == ChatType.PRIVATE:  # only send in PM's
            if ctx.input and ctx.input == "help":
                keyboard = await self.help_builder(chat.id)
                await ctx.respond(
                    await self.text(chat.id, "help-pm", self.bot_name),
                    reply_markup=InlineKeyboardMarkup(keyboard),
                )
                return None

            if ctx.input:
                rules_re = re.compile(r"rules_(.*)")
                if rules_re.search(ctx.input):
                    plug: "Rules" = self.bot.plugins["Rules"]  # type: ignore
                    try:
                        return await plug.start_rules(ctx)
                    except (ChannelInvalid, ChannelPrivate):
                        return await self.text(chat.id, "rules-channel-invalid")

                help_re = re.compile(r"help_(.*)").match(ctx.input)
                if help_re:
                    text_lang = await self.text(chat.id, f"{help_re.group(1)}-help")
                    text = (
                        f"Here is the help for the **{ctx.input.capitalize()}** "
                        f"plugin:\n\n{text_lang}"
                    )
                    await ctx.respond(
                        text,
                        reply_markup=InlineKeyboardMarkup(
                            [
                                [
                                    InlineKeyboardButton(
                                        await self.text(chat.id, "back-button"),
                                        callback_data="help_back",
                                    )
                                ]
                            ]
                        ),
                        parse_mode=ParseMode.MARKDOWN,
                    )
                    return

            permission = [
                "change_info",
                "post_messages",
                "edit_messages",
                "delete_messages",
                "restrict_members",
                "invite_users",
                "pin_messages",
                "promote_members",
                "manage_video_chats",
                "manage_chat",
            ]
            buttons = [
                [
                    InlineKeyboardButton(
                        text=await self.text(chat.id, "add-to-group-button"),
                        url=f"t.me/{self.bot.user.username}?startgroup=true&admin={'+'.join(permission)}",
                    ),
                    InlineKeyboardButton(
                        text=await self.text(chat.id, "start-help-button"),
                        url=f"t.me/{self.bot.user.username}?start=help",
                    ),
                ],
            ]
            if "Canonical" in self.bot.plugins:
                buttons.append(
                    [
                        InlineKeyboardButton(
                            text=await self.text(chat.id, "dashboard-button"),
                            url="https://userbotindo.com/dashboard",
                        )
                    ]
                )

            await ctx.respond(
                await self.text(chat.id, "start-pm", self.bot_name),
                reply_markup=InlineKeyboardMarkup(buttons),
                disable_web_page_preview=True,
                parse_mode=ParseMode.MARKDOWN,
            )
            return None

        return await self.text(chat.id, "start-chat")

    async def cmd_help(self, ctx: command.Context) -> None:
        """Bot plugins helper"""
        chat = ctx.chat

        if chat.type != ChatType.PRIVATE:  # only send in PM's
            await ctx.respond(
                await self.text(chat.id, "help-chat"),
                reply_markup=InlineKeyboardMarkup(
                    [
                        [
                            InlineKeyboardButton(
                                text=await self.text(chat.id, "help-chat-button"),
                                url=f"t.me/{self.bot.user.username}?start=help",
                            )
                        ]
                    ]
                ),
            )
            return

        keyboard = await self.help_builder(chat.id)
        await ctx.respond(
            await self.text(chat.id, "help-pm", self.bot_name),
            reply_markup=InlineKeyboardMarkup(keyboard),
        )

    async def cmd_donate(self, ctx: command.Context) -> None:
        """Bot donate command"""
        await ctx.respond(
            await self.text(ctx.chat.id, "donate"),
            disable_web_page_preview=True,
        )

    async def cmd_markdownhelp(self, ctx: command.Context) -> None:
        """Send markdown helper."""
        await ctx.respond(
            await self.text(ctx.chat.id, "markdown-helper", self.bot_name),
            parse_mode=ParseMode.HTML,
            disable_web_page_preview=True,
        )

    @command.filters(aliases=["fillinghelp"])
    async def cmd_formathelp(self, ctx: command.Context) -> None:
        """Send markdown help."""
        await ctx.respond(
            await self.text(ctx.chat.id, "filling-format-helper", noformat=True),
            parse_mode=ParseMode.MARKDOWN,
            disable_web_page_preview=True,
        )



"""Anjani database types"""
# Copyright (C) 2020 - 2023  UserbotIndo Team, <https://github.com/userbotindo.git>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import TypeVar, Union

from pymongo.operations import (
    DeleteMany,
    DeleteOne,
    InsertOne,
    ReplaceOne,
    UpdateMany,
    UpdateOne,
)
from pymongo.read_preferences import (
    Nearest,
    Primary,
    PrimaryPreferred,
    Secondary,
    SecondaryPreferred,
)

ReadPreferences = Union[Primary, PrimaryPreferred, Secondary, SecondaryPreferred, Nearest]
Request = TypeVar("Request", DeleteMany, DeleteOne, InsertOne, ReplaceOne, UpdateMany, UpdateOne)
Results = TypeVar("Results")
